#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/25 19:03
# @Author  : 兵
# @email    : 1747193328@qq.com

import math
import os.path

import numpy as np
from ase import Atoms
from ase.io import write as ase_write
from ase.io.vasp import read_vasp

from NepTrain import utils, Config, module_path
from NepTrain.core.utils import check_env


from .io import VaspInput,write_to_xyz

atoms_index=1

@utils.iter_path_to_atoms(["*.vasp","*.xyz"],show_progress=True,
                 description="VASP calculation progress" )
def calculate_vasp(atoms:Atoms,argparse):
    global atoms_index

    vasp = VaspInput()
    if argparse.incar is not None and os.path.exists(argparse.incar):
        vasp.read_incar(argparse.incar)
    else:
        vasp.read_incar(os.path.join(module_path,"core/dft/vasp/INCAR"))
    directory=os.path.join(argparse.directory,f"{atoms_index}-{atoms.get_chemical_formula()}")

    atoms_index+=1
    command = f"{Config.get('environ','mpirun_path')} -n {argparse.n_cpu} {Config.get('environ','vasp_path')}"
    if "NEPTRAIN_VASP_COMMAND" in os.environ:
        command = os.environ["NEPTRAIN_VASP_COMMAND"]

    a, b, c, alpha, beta, gamma = atoms.get_cell_lengths_and_angles()

    if argparse.kspacing is not None:
        vasp.set(kspacing=argparse.kspacing)
    vasp.set(
            directory = directory,
            command = command,
            kpts = (math.ceil(argparse.ka[0]/a)  ,
                  math.ceil(argparse.ka[1]/b)  ,
                  math.ceil(argparse.ka[2]/c) ),
            gamma = argparse.use_gamma,
             )

    if vasp.int_params["ibrion"] ==0:
        #分子动力学
        vasp.calculate(atoms, ('energy'))

        atoms_list = write_to_xyz(os.path.join(directory,"vasprun.xml"),os.path.join(directory,f"aimd_{vasp.float_params['tebeg']}k_{vasp.float_params['teend']}k.xyz"),"aimd",False)
        return atoms_list
    else:
        vasp.calculate(atoms, ('energy'))
        atoms.calc = vasp._xml_calc
        xx, yy, zz, yz, xz, xy = -vasp.results['stress'] * atoms.get_volume()  # *160.21766
        atoms.info['virial'] = np.array([(xx, xy, xz), (xy, yy, yz), (xz, yz, zz)])
        # 这里没想好怎么设计config的格式化  就先使用原来的
        if "Config_type" not in atoms.info:
            atoms.info['Config_type'] = "NepTrain scf "
        atoms.info['Weight'] = 1.0
        del atoms.calc.results['stress']
        del atoms.calc.results['free_energy']
        if vasp.converged:
            return atoms
        else:
            raise ValueError(f"{directory}: VASP not converged")
def run_vasp(argparse):
    check_env()

    result = calculate_vasp(argparse.model_path,argparse)
    path=os.path.dirname(argparse.out_file_path)
    if path and  not os.path.exists(path):
        os.makedirs(path)
    if len(result) and isinstance(result[0],list):
        result=[atoms for _list in result for atoms in _list]
    ase_write(argparse.out_file_path,result,format="extxyz",append=argparse.append)

    utils.print_success("VASP calculation task completed!" )

def set_magmom(directory):
  if 'magmom' in Config:
      items = Config.items('magmom')
      if items:
          element_magmoms = {}
          for symbol, moment_str in Config['magmom'].items():
              try:
                  element_magmoms[symbol] = float(moment_str.strip())
              except ValueError:
                  element_magmoms[symbol] = 0.0
          nonzero = False
          for value in element_magmoms.values():
              if value != 0.0:
                  nonzero = True
          if nonzero == True:
              poscar_path = os.path.join(directory, "POSCAR")
              atoms = read_vasp(poscar_path)
              symbols = atoms.get_chemical_symbols()
              unique_symbols_ordered = []
              seen_symbols = set()
              for symbol in symbols:
                  if symbol not in seen_symbols:
                      unique_symbols_ordered.append(symbol)
                      seen_symbols.add(symbol)
          
              symbol_counts = {symbol: symbols.count(symbol) for symbol in symbols}
          
              magmom_lines = []
              for symbol in unique_symbols_ordered:
                  count = symbol_counts[symbol]
                  try:
                      magmom_lines.append(f"{element_magmoms[symbol]}*{count}")
                  except:
                      magmom_lines.append(f"0.0*{count}")
          
              magmom_string = " ".join(magmom_lines)
              magmom_line = f"{magmom_string}\n"
              return magmom_line
          else:
              return None
      else:
          return None
  else:
      return None

if __name__ == '__main__':
    calculate_vasp("./")



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2025/5/9 19:41
# @Author  : 兵
# @email    : 1747193328@qq.com
import numpy as np
from ase import Atoms
from ase.data import atomic_numbers,covalent_radii,chemical_symbols
from joblib import Parallel, delayed
from tqdm import tqdm

def calculate_pairwise_distances(lattice_params:np.ndarray, atom_coords:np.ndarray, fractional=True):
    """
    计算晶体中所有原子对之间的距离，考虑周期性边界条件

    参数:
    lattice_params: 晶格参数，3x3 numpy array 表示晶格向量 (a, b, c)
    atom_coords: 原子坐标，Nx3 numpy array
    fractional: 是否为分数坐标 (True) 或笛卡尔坐标 (False)

    返回:
    distances: NxN numpy array，所有原子对之间的距离
    """


    if fractional:
        atom_coords = np.dot(atom_coords, lattice_params)

    diff = atom_coords[np.newaxis, :, :] - atom_coords[:, np.newaxis, :]
    shifts = np.array(np.meshgrid([-1, 0, 1], [-1, 0, 1], [-1, 0, 1]), dtype=np.int8).T.reshape(-1, 3)
    lattice_shifts = np.dot(shifts, lattice_params)
    all_diffs = diff[:, :, np.newaxis, :] + lattice_shifts[np.newaxis, np.newaxis, :, :]
    all_distances = np.sqrt(np.sum(all_diffs ** 2, axis=-1))
    distances = np.min(all_distances, axis=-1)
    np.fill_diagonal(distances, 0)
    return distances


def get_mini_distance_info(atoms:Atoms):
    """
    返回原子对之间的最小距离
    """
    dist_matrix = calculate_pairwise_distances(atoms.cell, atoms.positions, False)
    symbols = atoms.get_chemical_symbols()
    # 提取上三角矩阵（排除对角线）
    i, j = np.triu_indices(len(atoms), k=1)
    # 用字典来存储每种元素对的最小键长
    bond_lengths = {}
    # 遍历所有原子对，计算每一对元素的最小键长
    for idx in range(len(i)):
        atom_i, atom_j = symbols[i[idx]], symbols[j[idx]]
        # if atom_i==atom_j:
        #     continue
        # 获取当前键长
        bond_length = dist_matrix[i[idx], j[idx]]
        # if bond_length>5:
        #     continue
        # 确保元素对按字母顺序排列，避免 Cs-Ag 和 Ag-Cs 视为不同
        element_pair = tuple(sorted([atom_i, atom_j]))
        # 如果该元素对尚未存在于字典中，初始化其最小键长
        if element_pair not in bond_lengths:
            bond_lengths[element_pair] = bond_length
        else:
            # 更新最小键长
            bond_lengths[element_pair] = min(bond_lengths[element_pair], bond_length)

    return bond_lengths

def adjust_reasonable(atoms, coefficient=0.7):
    """
    根据传入系数 对比共价半径和实际键长，
    如果实际键长小于coefficient*共价半径之和，判定为不合理结构 返回False
    否则返回 True
    :param coefficient: 系数
    :return:

    """
    distance_info = get_mini_distance_info(atoms)
    for elems, bond_length in distance_info.items():

        r1=covalent_radii[chemical_symbols.index(elems[0])]
        r2=covalent_radii[chemical_symbols.index(elems[1])]
        if isinstance(r1,float) and isinstance(r2,float):

            # 相邻原子距离小于共价半径之和×系数就选中
            if (r1 + r2) * coefficient > bond_length :
                return False
    return True



def process_atom(atoms, filter_func):
    if adjust_reasonable(atoms, filter_func):
        return atoms, True
    else:
        return atoms, False

def parallel_filter_trajectory(trajectory, filter_func, n_jobs=-1):
    """使用 joblib 并行处理"""
    results = Parallel(n_jobs=n_jobs)(
        delayed(process_atom)(atoms, filter_func)
        for atoms in tqdm(trajectory, desc="Filtering structures")
    )

    trajectory_structures = [atoms for atoms, is_reasonable in results if is_reasonable]
    filter_structures = [atoms for atoms, is_reasonable in results if not is_reasonable]

    return trajectory_structures, filter_structures





#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2025/8/2 11:07
# @Author  : 兵
# @email    : 1747193328@qq.com
from .vasp import run_vasp
def run_dft(argparse):
    # 默认是 vasp

    if argparse.software is None:

        argparse.software = 'vasp'


    if argparse.directory is None:
        argparse.directory = f"./cache/{argparse.software}"
    if argparse.out_file_path is None:
        argparse.out_file_path = f"./{argparse.software}_scf.xyz"

    if argparse.software =="vasp":

        if argparse.incar is None:
            argparse.incar=f"./INCAR"

        return run_vasp(argparse)
    elif argparse.software =="abacus":
        from .abacus import run_abacus

        if argparse.incar is None:
            argparse.incar=f"./INPUT"
        return run_abacus(argparse)


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/25 13:37
# @Author  : 兵
# @email    : 1747193328@qq.com
"""
自动训练的逻辑
"""
import os.path
import shutil
import subprocess
import asyncio
from pathlib import Path
from typing import List, Tuple
from ase.io import read as ase_read
from ase.io import write as ase_write
from .worker import submit_job, async_submit_job
from ruamel.yaml import YAML

from NepTrain import utils

from ..utils import check_env


async def _await_tasks(tasks):
    """Helper coroutine to run multiple async tasks."""
    await asyncio.gather(*tasks)

def filter_file_path(forward_files,base_dir=""):
    new_files = []
    for forward_file in forward_files:
        if os.path.exists(os.path.join(base_dir,forward_file)):
            new_files.append(forward_file)
    return new_files
def relpath_from_files(files, start):
    if isinstance(files, (list, tuple)):
        return [os.path.relpath(file,start) for file in files]
    return os.path.relpath(files, start)
PARAMS = Tuple[str,list ]



class Manager:
    def __init__(self, options):
        self.options = options
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self.options):
            self.index = 0
        value = self.options[self.index]
        self.index += 1
        return value

    def set_next(self, option):
        index=self.options.index(option)
        # 设置当前索引，注意索引从0开始
        if 0 <= index < len(self.options):
            self.index = index
        else:
            raise IndexError("Index out of range.")


class PathManager:



    def __init__(self, root):
        self.root = root

    def __getattr__(self, item):
        return os.path.join(self.root, item)

def params2str(params):
    text=""
    for i in  params:
        if isinstance(i, str):
            text += i
        elif isinstance(i, (tuple,list)):
            for j in i:
                text += str(j)
                text += " "
        else:
            text += str(i)
        text += " "
    # print(text)
    return text

class NepTrainWorker:
    pass
    def __init__(self):
        self.config={}
        self.job_list=["nep","gpumd","select","dft","pred", ]
        self.manager=Manager(self.job_list)



    def __getattr__(self, item):

        if item.startswith("last_"):
            item=item.replace("last_","")
            generation_path=os.path.join((self.config.get("work_path")), f"Generation-{self.generation-1}")
        else:
            generation_path=os.path.join((self.config.get("work_path")), f"Generation-{self.generation}")

        if item=="generation_path":

            return generation_path

        items= item.split("_")
        if items[0] in self.job_list:
            job_path=os.path.join(generation_path, items.pop(0))
        else:
            job_path=generation_path
        fin_path=os.path.join(job_path, "_".join(items[:-1]) )
        if items[-1]=="path":
            pass
            utils.verify_path(fin_path)
        else:
            last_underscore_index = fin_path.rfind('_')
            if last_underscore_index != -1:
                # 替换最后一个下划线为点
                fin_path = fin_path[:last_underscore_index] + '.' + fin_path[last_underscore_index + 1:]
            else:
                fin_path = fin_path

            utils.verify_path(os.path.dirname(fin_path))


        return fin_path



    @property
    def generation(self):
        return self.config.get("generation")
    @generation.setter
    def generation(self,value):
        self.config["generation"] = value



    def split_dft_job_xyz(self,xyz_file):
        addxyz = ase_read(xyz_file, ":", format="extxyz")

        split_addxyz_list = utils.split_list(addxyz, self.config["dft_job"])


        for i, xyz in enumerate(split_addxyz_list):
            if xyz:
                ase_write(self.__getattr__(f"dft_learn_add_{i + 1}_xyz_file"), xyz, format="extxyz")

    def check_env(self):


        if self.config.get("restart") :
            utils.print("No need for initialization check.")
            utils.print_msg("--" * 4,
                            f"Restarting to train the potential function for the {self.generation}th generation.",
                            "--" * 4)

            return

        if self.config["current_job"]=="dft":

            self.generation=0
            utils.copy(self.config["init_train_xyz"], self.select_selected_xyz_file)

            # if self.config["dft_job"] != 1:
            #
            #
            #     self.split_dft_job_xyz(self.config["init_train_xyz"])
        elif self.config["current_job"]=="nep":
           

            utils.copy(self.config["init_train_xyz"], self.last_all_learn_calculated_xyz_file)
            # utils.copy(self.config["init_train_xyz"], self.last_all_learn_calculated_xyz_file )
            #如果势函数有效  直接先复制过来
        elif self.config["current_job"]=="gpumd":

            utils.copy(self.config["init_train_xyz"],self.nep_train_xyz_file )

            if os.path.exists(self.config["init_nep_txt"]):
                utils.copy(self.config["init_nep_txt"],
                            self.nep_nep_txt_file )
            else:
                raise FileNotFoundError("Starting task as gpumd requires specifying a valid potential function path!")
        else:
            raise ValueError("current_job can only be one of nep, gpumd, or dft.")

    def read_config(self,config_path):
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"The file at {config_path} does not exist.")
        with open(config_path,"r",encoding="utf8") as f:


            self.config=YAML().load(f )
        if self.config["dft"]["incar_path"]=="auto":
            if self.config["dft"]["software"]=="abacus":
                self.config["dft"]["incar_path"]="./INPUT"
            else:
                self.config["dft"]["incar_path"]="./INCAR"
            
    def build_pred_params(self):
        nep=self.config["nep"]

        utils.copy(nep.get("nep_in_path"), self.pred_nep_in_file)
        utils.copy(self.nep_nep_txt_file, self.pred_nep_txt_file)
        utils.copy(self.all_learn_calculated_xyz_file, self.pred_train_xyz_file)

        params=[]
        params.append("NepTrain")
        params.append("nep")

        params.append("--directory")
        params.append("./")

        params.append("--in")
        params.append("nep.in")

        params.append("--train")
        params.append("train.xyz")

        params.append("--nep")
        params.append("nep.txt")

        params.append("--prediction")

        return params2str(params)


    def build_nep_params(self) :
        nep=self.config["nep"]

        utils.copy(self.last_improved_train_xyz_file, self.nep_train_xyz_file)
        utils.copy(nep.get("nep_in_path"), self.nep_nep_in_file)
        utils.copy(nep.get("test"), self.nep_test_xyz_file)
        utils.copy(self.last_nep_nep_restart_file, self.nep_nep_restart_file)

        params=[]
        params.append("NepTrain")
        params.append("nep")

        params.append("--directory")
        params.append("./")

        params.append("--in")
        params.append("nep.in")


        params.append("--train")
        params.append("train.xyz")

        # params.append(relpath_from_files(self.last_improved_train_xyz_file,self.nep_path))

        params.append("--test")
        params.append("test.xyz")
        #
        # params.append(relpath_from_files(nep.get("test_xyz_path"),self.nep_path))

        if self.config["nep"]["nep_restart"] and self.generation not in [1,len(self.config["gpumd"]["step_times"])+1]:
            #开启续跑
            #如果上一级的势函数路径有效  就传入一下续跑的参数

            if os.path.exists(self.last_nep_nep_restart_file):
                utils.print_tip("Start the restart mode!")
                params.append("--restart_file")
                params.append("nep.restart")
                # params.append(relpath_from_files(self.last_nep_nep_restart_file,self.nep_path))
                params.append("--continue_step")
                params.append(self.config["nep"]["nep_restart_step"])

        return params2str(params)
    def build_gpumd_params(self,model_path,temperature,n_job=1,):
        gpumd=self.config["gpumd"]
        base_name = os.path.basename(model_path)
        utils.copy(model_path, os.path.join(self.gpumd_path, base_name))

        utils.copy(gpumd.get("run_in_path"), self.gpumd_run_in_file)

        utils.copy(self.nep_nep_txt_file, self.gpumd_nep_txt_file)

        params=[]
        params.append("NepTrain")
        params.append("gpumd")

        params.append(base_name)

        params.append("--directory")

        params.append("./")

        params.append("--in")
        params.append("run.in")
        params.append("--nep")
        params.append( "nep.txt")
        params.append("--time")
        params.append(gpumd.get("step_times")[self.generation-1])

        params.append("--temperature")

        params.append(temperature)




        params.append("--out")
        params.append( f"./trajectory_{n_job}.xyz")




        return params2str(params)
    def build_select_params(self):
        select=self.config["select"]
        utils.copy(self.nep_nep_txt_file, self.select_nep_txt_file)

        utils.copy(self.nep_train_xyz_file, self.select_train_xyz_file)

        params=[]
        params.append("NepTrain")
        params.append("select")
        #总的
        params.append("trajectorys.xyz")
        #分开
        # params.append(relpath_from_files(self.__getattr__(f"select_md_*_xyz_file"),self.select_path ))
        params.append("--nep")
        params.append( relpath_from_files(self.select_nep_txt_file,self.select_path ))

        params.append("--base")
        params.append( relpath_from_files(self.select_train_xyz_file ,self.select_path ))
        params.append("--max_selected")
        params.append(select["max_selected"])
        params.append("--min_distance")
        params.append(select["min_distance"])
        params.append("--out")
        params.append(relpath_from_files(self.select_selected_xyz_file,self.select_path ))

        if select.get("filter",False):

            params.append("--filter")
            params.append(select.get("filter" ) if isinstance(select.get("filter" ),float) else 0.6)


        return params2str(params)


    def build_dft_params(self,n_job=1):
        dft=self.config["dft"]

        utils.copy(dft["incar_path"], self.dft_path)

        params=[]
        params.append("NepTrain")
        params.append("dft")

        if self.config["dft_job"] == 1:

            if not os.path.exists(self.dft_learn_add_xyz_file):
                return None
            params.append(relpath_from_files(self.dft_learn_add_xyz_file,self.dft_path ))
        else:
            path=self.__getattr__(f"dft_learn_add_{n_job}_xyz_file")
            if not os.path.exists(path):
                return None
            params.append(relpath_from_files(path,self.dft_path ))

        params.append("--directory")

        params.append(relpath_from_files(self.__getattr__(f"dft_cache{n_job}_path"),self.dft_path ))


        params.append("-np")
        params.append(dft["cpu_core"])
        if dft["kpoints_use_gamma"]:
            params.append("--gamma")

        if dft["incar_path"]:

            params.append("--in")

            params.append(os.path.basename(dft["incar_path"]))
        if dft["use_k_stype"] == "kpoints":
            if dft.get("kpoints"):
                params.append("-ka")
                if isinstance(dft["kpoints"],list):
                    params.append(",".join([str(i) for i in dft["kpoints"]]))
                else:
                    params.append(dft["kpoints"])
        else:

            if dft.get("kspacing") :
                params.append("--kspacing")
                params.append(dft["kspacing"])
        # params.append("--software")
        params.append("--" + dft["software"])
        params.append("--out")
        params.append( relpath_from_files(self.__getattr__(f"dft_learn_calculated_{n_job}_xyz_file"),self.dft_path ))


        return params2str(params)
    def sub_select(self):
        # utils.cat(self.__getattr__(f"select_md_*_xyz_file"),
        #           self.select_all_md_dummp_xyz_file
        #           )
        utils.print_msg(f"Start sampling from the trajectory.")
        utils.cat(self.__getattr__(f"gpumd_trajectory_*_xyz_file"),
                  self.select_trajectorys_xyz_file
                  )

        if utils.is_file_empty(self.select_trajectorys_xyz_file):
            utils.print_warning(f"No trajectory file, skip sampling")

            return

        cmd = self.build_select_params()


        submit_job(
            machine_dict=self.config["select"]["machine"],
            resources_dict=self.config["select"]["resources"],
            task_dict_list=[
                dict(
                    command=cmd,
                    task_work_path="./",
                    forward_files=filter_file_path(["nep.txt", "train.xyz","trajectorys.xyz"]),
                    backward_files=relpath_from_files([
                        self.select_selected_xyz_file,
                        self.select_selected_png_file,
                                    self.__getattr__(f"select_selected_md_*_*_file")
                                    ],self.select_path),
                )
            ],
            submission_dict=dict(
                work_base=self.select_path,
                forward_common_files=[],
                backward_common_files=[],

            )

        )








    def sub_dft(self):
        utils.print_msg("Beginning the execution of VASP for single-point energy calculations.")
        # break
        utils.cat(self.select_selected_xyz_file,
                  self.dft_learn_add_xyz_file
                  )

        if not utils.is_file_empty(self.dft_learn_add_xyz_file):
            if self.config["dft"]["software"] == "abacus":
                from NepTrain.core.dft.abacus import StructureVar
                StructureVar.init(self.config["gpumd"]["model_path"])
                StructureVar.init("./")
                for pp in StructureVar.pp_files.values():
                    curr_p=f"./{pp}"
                    stru_p=f'{self.config["gpumd"]["model_path"]}/{pp}'
                    if os.path.exists(curr_p):
                        shutil.copy(curr_p,self.dft_path)
                    elif os.path.exists(stru_p):
                        shutil.copy(stru_p,self.dft_path)
                for orb in StructureVar.orbs.values():
                    curr_orb=f"./{orb}"
                    stru_orb=f'{self.config["gpumd"]["model_path"]}/{orb}'
                    if os.path.exists(curr_orb):
                        shutil.copy(curr_orb,self.dft_path)
                    elif os.path.exists(stru_orb):
                        shutil.copy(stru_orb,self.dft_path)
            if self.config["dft_job"] != 1:
                # Split xyz for parallel submission
                self.split_dft_job_xyz(self.dft_learn_add_xyz_file)

            tasks = []
            for i in range(self.config["dft_job"]):
                cmd = self.build_dft_params(i + 1)
                if cmd is None:
                    continue
                if self.config["dft_job"] == 1:
                    forward_files=["learn_add.xyz",os.path.basename(self.config["dft"]["incar_path"])]
                else:
                    forward_files=[f"learn_add_{i + 1}.xyz",os.path.basename(self.config["dft"]["incar_path"])]

                tasks.append(
                    async_submit_job(
                        machine_dict=self.config["dft"]["machine"],
                        resources_dict=self.config["dft"]["resources"],
                        task_dict_list=[
                            dict(
                                command=cmd,
                                task_work_path="./",
                                forward_files=filter_file_path(forward_files),
                                backward_files=[f"learn_calculated_{i +1}.xyz"],
                            )
                        ],
                        submission_dict=dict(
                            work_base=self.dft_path,
                            forward_common_files=[],
                            backward_common_files=[],
                        ),
                    )
                )

            if tasks:
                asyncio.run(_await_tasks(tasks))

            utils.cat(self.__getattr__(f"dft_learn_calculated_*_xyz_file"),
                      self.all_learn_calculated_xyz_file
                      )
            if self.config.get("limit",{}).get("force") and not utils.is_file_empty(self.all_learn_calculated_xyz_file):
                bad_structure = []
                good_structure = []
                structures=ase_read(self.all_learn_calculated_xyz_file,":")
                for structure in structures:

                    if structure.calc.results["forces"].max() <= self.config.get("limit",{}).get("force"):
                        good_structure.append(structure)
                    else:
                        bad_structure.append(structure)

                ase_write(self.all_learn_calculated_xyz_file,good_structure,append=False,format="extxyz")
                if bad_structure:
                    ase_write(self.remove_by_force_xyz_file, bad_structure, append=False, format="extxyz")

        else:
            utils.print_warning("Detected that the calculation input file is empty, proceeding directly to the next step!")

            utils.cat(self.dft_learn_add_xyz_file,
                      self.all_learn_calculated_xyz_file
                      )

    def sub_nep(self):
        utils.print_msg("--" * 4, f"Starting to train the potential function for the {self.generation}th generation.", "--" * 4)

        if not utils.is_file_empty(self.last_all_learn_calculated_xyz_file):


            if os.path.exists(self.last_nep_train_xyz_file):
                utils.cat([self.last_nep_train_xyz_file,
                           self.last_all_learn_calculated_xyz_file
                           ],
                          self.last_improved_train_xyz_file

                          )
            else:
                utils.copy(self.last_all_learn_calculated_xyz_file,
                            self.last_improved_train_xyz_file)

            utils.print_msg(f"Starting to train the potential function.")
            cmd = self.build_nep_params()


            submit_job(
                machine_dict = self.config["nep"]["machine"],
                resources_dict = self.config["nep"]["resources"],
                task_dict_list = [
                    dict(
                        command=cmd,
                        task_work_path="./",
                        forward_files= filter_file_path(["nep.in","nep.restart","train.xyz","test.xyz"]),
                        backward_files = [
                            "./*"
                        ],
                    )
                ],
                submission_dict = dict(
                    work_base=self.nep_path,
                    forward_common_files=[],
                    backward_common_files=[],

                )

            )

        else:
            utils.print_warning("The dataset has not changed, directly copying the potential function from the last time!")

            utils.copy_files(self.last_nep_path, self.nep_path)

    def sub_nep_pred(self):

        if utils.is_file_empty(self.nep_nep_txt_file):
            utils.print_msg(f"No potential function available, skipping prediction.")
            return
        if not utils.is_file_empty(self.all_learn_calculated_xyz_file):
            utils.print_msg(f"Starting to predict new dataset.")
            cmd = self.build_pred_params()
            submit_job(
                machine_dict=self.config["nep"]["machine"],
                resources_dict=self.config["nep"]["resources"],
                task_dict_list=[
                    dict(
                        command=cmd,
                        task_work_path="./",
                        forward_files=filter_file_path(["nep.in","nep.txt","train.xyz"]),
                        backward_files=["./*"],
                    )
                ],
                submission_dict=dict(
                    work_base=self.pred_path,
                    forward_common_files=[],
                    backward_common_files=[],
                ),
            )
        else:
            utils.print_msg(f"The dataset has not changed, skipping prediction.")


    def sub_gpumd(self):


        utils.print_msg(f"Starting active learning.")
        tasks = []
        if self.config.get("gpumd_split_job", "temperature") == "temperature":
            for i, temp in enumerate(self.config["gpumd"]["temperature_every_step"]):
                cmd = self.build_gpumd_params(
                    self.config["gpumd"].get("model_path"),
                    temp,
                    i,
                )
                base_name=os.path.basename(self.config["gpumd"].get("model_path"))
                tasks.append(
                    async_submit_job(
                        machine_dict=self.config["gpumd"]["machine"],
                        resources_dict=self.config["gpumd"]["resources"],
                        task_dict_list=[
                            dict(
                                command=cmd,
                                task_work_path="./",
                                forward_files=filter_file_path(["run.in","nep.txt",base_name]),
                                backward_files=[f"./trajectory_{i}.xyz"],
                            )
                        ],
                        submission_dict=dict(
                            work_base=self.gpumd_path,
                            forward_common_files=[],
                            backward_common_files=[],
                        ),
                    )
                )
        else:
            if os.path.isdir(self.config["gpumd"]["model_path"]):
                for i, file in enumerate(os.listdir(self.config["gpumd"]["model_path"])):
                    cmd = self.build_gpumd_params(
                        os.path.join(self.config["gpumd"]["model_path"], file),
                        self.config["gpumd"]["temperature_every_step"],
                        i,
                    )
                    tasks.append(
                        async_submit_job(
                            machine_dict=self.config["gpumd"]["machine"],
                            resources_dict=self.config["gpumd"]["resources"],
                            task_dict_list=[
                                dict(
                                    command=cmd,
                                    task_work_path="./",
                                    forward_files=["run.in","nep.txt",file],
                                    backward_files=[f"./trajectory_{i}.xyz"],
                                )
                            ],
                            submission_dict=dict(
                                work_base= self.gpumd_path,
                                forward_common_files=[],
                                backward_common_files=[],
                            ),
                        )
                    )
        if tasks:
            asyncio.run(_await_tasks(tasks))

        # utils.cat(self.__getattr__(f"gpumd_trajectory_*_xyz_file"),
        #           self.select_trajectorys_xyz_file
        #           )


    def start(self,config_path):
        utils.print_msg("Welcome to NepTrain automatic training!")

        self.read_config(config_path)
        self.check_env()



        self.manager.set_next(self.config.get("current_job"))

        while True:

            #开始循环
            job = next(self.manager)
            # utils.print_msg(f"[Generation {self.generation}] Starting job: {job}")
            self.config["current_job"]=job
            self.save_restart()
            if job=="dft":

                self.sub_dft()

            elif job=="pred":

                self.sub_nep_pred()
                self.generation += 1

            elif job=="nep":

                self.sub_nep()
                if self.generation>len(self.config["gpumd"]["step_times"]):
                   utils.print_success("Training completed!")
                   break
            elif job=="select":

                self.sub_select()

            else:
                if utils.is_file_empty(self.nep_nep_txt_file):
                    utils.print_warning(f"No potential function available, break!!!")
                    break
                self.sub_gpumd()

            # utils.print_msg(f"[Generation {self.generation}] Finished job: {job}")


    def save_restart(self):
        with open("./restart.yaml","w",encoding="utf-8") as f:
            self.config["restart"]=True

            YAML().dump(self.config,f)

def train_nep(argparse):
    """
    首先检查下当前的进度 看从哪开始
    :return:
    """


    worker = NepTrainWorker()

    worker.start(argparse.config_path)
if __name__ == '__main__':
    train =NepTrainWorker()
    train.generation=1
    train.config["work_path"]="./cache"
    print(train.nep_path)

    print(train.__getattr__(f"dft_learn_calculated_*_xyz_file"))



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/21 14:22
# @Author  : 兵
# @email    : 1747193328@qq.com
import contextlib
import os

import numpy as np
from ase import Atoms

from NepTrain.nep_cpu import CpuNep

class Nep3Calculator:

    def __init__(self, model_file="nep.txt"):
        if not isinstance(model_file, str):
            model_file=str(model_file,encoding="utf-8")
        with open(os.devnull, 'w') as devnull:
            with contextlib.redirect_stdout(devnull), contextlib.redirect_stderr(devnull):
                self.nep3 = CpuNep(model_file)
        self.element_list=self.nep3.get_element_list()
        self.type_dict = {e: i for i, e in enumerate(self.element_list)}




    def get_descriptors(self,structure):
        symbols = structure.get_chemical_symbols()
        _type = [self.type_dict[k] for k in symbols]
        _box = structure.cell.transpose(1, 0).reshape(-1).tolist()
        _position = structure.get_positions().transpose(1, 0).reshape(-1).tolist()
        descriptor = self.nep3.get_descriptor(_type, _box, _position)
        descriptors_per_atom = np.array(descriptor).reshape(-1, len(structure)).T

        return descriptors_per_atom
    def get_structure_descriptors(self, structure):
        descriptors_per_atom=self.get_descriptors(structure)
        return descriptors_per_atom.mean(axis=0)

    def get_structures_descriptors(self,structures:[Atoms]):
        _types=[]
        _boxs=[]
        _positions=[]

        for structure in structures:
            symbols = structure.get_chemical_symbols()
            _type = [self.type_dict[k] for k in symbols]
            _box = structure.cell.transpose(1, 0).reshape(-1).tolist()
            _position = structure.get_positions().transpose(1, 0).reshape(-1).tolist()
            _types.append(_type)
            _boxs.append(_box)
            _positions.append(_position)

        descriptor = self.nep3.get_descriptors(_types, _boxs, _positions)

        return np.array(descriptor)


    def calculate(self,structures:list[Atoms]):
        group_size=[]
        _types = []
        _boxs = []
        _positions = []
        if isinstance(structures, Atoms):
            structures = [structures]
        for structure in structures:
            symbols = structure.get_chemical_symbols()
            _type = [self.type_dict[k] for k in symbols]
            _box = structure.cell.transpose(1, 0).reshape(-1).tolist()
            _position = structure.get_positions().transpose(1, 0).reshape(-1).tolist()
            _types.append(_type)
            _boxs.append(_box)
            _positions.append(_position)
            group_size.append(len(_type))

        potentials, forces, virials = self.nep3.calculate(_types, _boxs, _positions)


        split_indices = np.cumsum(group_size)[:-1]
        #
        potentials=np.hstack(potentials)
        split_potential_arrays = np.split(potentials, split_indices)
        potentials_array = np.array(list(map(np.sum, split_potential_arrays)))
        # print(potentials_array)

        # 处理每个force数组：reshape (3, -1) 和 transpose(1, 0)
        reshaped_forces = [np.array(force).reshape(3, -1).T for force in forces]

        forces_array = np.vstack(reshaped_forces)
        # print(forces_array)

        reshaped_virials = np.vstack([np.array(virial).reshape(9, -1).mean(axis=1) for virial in virials])

        # virials_array = reshaped_virials[:,[0,4,8,1,5,6]]
        virials_array=reshaped_virials
        return potentials_array,forces_array,virials_array


class DescriptorCalculator:
    def __init__(self, calculator_type="nep",**calculator_kwargs):
        self.calculator_type=calculator_type
        if calculator_type == "nep":
            self.calculator=Nep3Calculator(**calculator_kwargs)
        elif calculator_type == "soap":
            from dscribe.descriptors import SOAP

            self.calculator = SOAP(
                **calculator_kwargs,dtype="float32"
            )
        else:
            raise ValueError("calculator_type must be nep or soap")


    def get_structures_descriptors(self,structures:[Atoms]):

        if len(structures)==0:
            return np.array([])

        if self.calculator_type == "nep":
            return self.calculator.get_structures_descriptors(structures)
        else:

            return  np.array([self.calculator.create_single(structure).mean(0) for structure in structures])


if __name__ == '__main__':
    nep3 = Nep3Calculator(model_file="/mnt/d/Desktop/vispy/KNbO3/nep.txt")
    from ase.io import read
    import time
    structures = read("/mnt/d/Desktop/vispy/KNbO3/train.xyz",index=":")
    start=time.time()

    descriptors = nep3.get_structures_descriptors(structures)
    print(f"计算描述符：{len(structures)}个结构，耗时：{time.time()-start:.3f}s")
    print("descriptors",descriptors.shape)
    start=time.time()

    potentials ,forces ,virials   = nep3.calculate(structures)

    print(f"计算性质：{len(structures)}个结构，耗时：{time.time()-start:.3f}s")
    print("potentials",potentials.shape)
    print("forces",forces.shape)
    print("virials",virials.shape)




#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/29 15:06
# @Author  : 兵
# @email    : 1747193328@qq.com

import os.path
import shutil

import numpy as np
from ase import Atoms
from ase.io import read as ase_read
from ase.io import write as ase_write

from NepTrain import utils, module_path

from ..select import select_structures, filter_by_bonds


from .io import RunInput


from ..utils import check_env

import zlib

from ...utils import print


def array_to_id(arr):
    arr = np.ascontiguousarray(arr)  # 确保内存连续
    return zlib.crc32(arr.tobytes())


@utils.iter_path_to_atoms(["*.vasp","*.xyz"],show_progress=False)
def calculate_gpumd(atoms:Atoms,argparse):

    atoms_index = array_to_id(atoms.positions)

    new_atoms=[]

    # if os.path.exists(argparse.out_file_path):
    #     os.remove(argparse.out_file_path)
    for temperature in argparse.temperature:

        run = RunInput(argparse.nep_txt_path)
        if utils.is_file_empty(argparse.run_in_path):
            run_in_path=os.path.join(module_path,"core/gpumd/run.in")
        else:
            run_in_path=argparse.run_in_path
        run.read_run(run_in_path)
        run.set_time_temp(argparse.time,temperature)
        directory=os.path.join(argparse.directory,f"{atoms_index}-{atoms.symbols}@{temperature}k-{argparse.time}ps")
        utils.print_msg(f"GPUMD is running, temperature: {temperature}k. Time: {argparse.time}ps" )

        run.calculate(atoms,directory)

        dump = ase_read(os.path.join(directory,"dump.xyz"), ":", format="extxyz", do_not_split_by_at_sign=True)
        for i, atom in enumerate(dump):
            atom.info["Config_type"] = f"{atom.symbols}-epoch-{argparse.time}ps-{temperature}k-{i + 1}"


        ase_write(argparse.out_file_path,dump,append=True)







    return new_atoms
def run_gpumd(argparse):
    check_env()
    utils.verify_path(os.path.dirname(os.path.abspath(argparse.out_file_path)))
    result = calculate_gpumd(argparse.model_path,argparse)












#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 14:22
# @Author  : 兵
# @email    : 1747193328@qq.com
"""Pymatgen package configuration."""

from __future__ import annotations

import platform
import sys
 
import os
import subprocess
import pybind11
from pybind11.setup_helpers import Pybind11Extension
from setuptools import  Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

# 获取 pybind11 的 include 路径
pybind11_include = pybind11.get_include()


# 设定编译选项

extra_link_args = ["-O3","-std=c++11"]
extra_compile_args=["-O3","-std=c++11"]
# 检查平台并设置相应的 OpenMP 编译标志

if sys.platform == "win32":
    # 对于 Windows 使用 MSVC 编译器时，需要使用 /openmp
    extra_compile_args.append('/openmp')


elif sys.platform == "darwin":
    # 对于 macOS 和 Clang 使用 -fopenmp 编译标志

    extra_compile_args.append('-fopenmp' )

    extra_link_args.append('-fopenmp')


else:
    # 对于 Linux 和 GCC 使用 -fopenmp 编译标志

    extra_compile_args.append('-fopenmp' )

    extra_link_args.append('-fopenmp')


# 定义扩展模块
ext_modules = [
    Pybind11Extension(
        "NepTrain.nep_cpu",  # 模块名
        ["src/nep_cpu/nep_cpu.cpp"],  # 源文件
        include_dirs=[
            pybind11_include,
            # "src/nep_cpu"
        ],
        extra_compile_args=extra_compile_args,  # 编译选项
        extra_link_args=extra_link_args,
        language="c++",  # 指定语言为 C++
    ),
]

# 自定义 build_ext 命令，确保兼容性
class BuildExt(build_ext):
    def build_extensions(self):
        # 设置编译器标准为 C++17
        ct = self.compiler.compiler_type
        opts = [ ]
        for ext in self.extensions:
            ext.extra_compile_args = opts + ext.extra_compile_args
        build_ext.build_extensions(self)



setup(
    author="Chen Cheng bing",
cmdclass={'build_ext': BuildExt},
    # include_dirs=[np.get_include()],
packages=find_packages(),
ext_modules=ext_modules,
zip_safe=False,
include_package_data=True,  # 确保包含额外的文件（如 .so）
)


import asyncio
import os

from dpdispatcher import Machine, Resources, Task, Submission
from pathlib import Path
from NepTrain import utils


def remove_sub_file(work_path: str = "./"):
    """Remove temporary submission files."""
    all_file = Path(work_path).glob("????????????????????????????????????????.sub.*")
    for file in all_file:
        file.unlink()
    all_file = Path(work_path).glob("????????????????????????????????????????.sub")
    for file in all_file:
        file.unlink()

def submit_job(
    machine_dict: dict,
    resources_dict: dict,
    task_dict_list: list,
    submission_dict: dict,
) -> Submission:
    """Submit a job synchronously using dpdispatcher."""
    machine = Machine.load_from_dict(machine_dict)
    resources = Resources.load_from_dict(resources_dict)
    task_list = [Task(**task_dict) for task_dict in task_dict_list]
    submission = Submission(
        machine=machine,
        resources=resources,
        task_list=task_list,
        **submission_dict,
    )
    submission.run_submission(clean=False)
    for job in submission.belonging_jobs:
        utils.print_msg(f"Finished job {job.job_id} in {job.machine.context.remote_root}")


    return submission


async def async_submit_job(
    machine_dict: dict,
    resources_dict: dict,
    task_dict_list: list,
    submission_dict: dict,
) -> None:
    """Submit a job asynchronously using dpdispatcher."""
    machine = Machine.load_from_dict(machine_dict)
    resources = Resources.load_from_dict(resources_dict)
    task_list = [Task(**task_dict) for task_dict in task_dict_list]
    submission = Submission(
        machine=machine,
        resources=resources,
        task_list=task_list,
        **submission_dict,
    )
    await submission.async_run_submission(check_interval=1, clean=False)
    for job in submission.belonging_jobs:
        utils.print_msg(f"Finished job {job.job_id} in {job.machine.context.remote_root}")





#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 15:53
# @Author  : 兵
# @email    : 1747193328@qq.com
import os
import subprocess
from pathlib import Path

import numpy as np
from ase.calculators import calculator
from ase.calculators.calculator import Calculator
from ase.calculators.vasp import Vasp
from ase.calculators.vasp.vasp import check_atoms
from ase.io import read as ase_read
from ase.io import write as ase_write
from NepTrain import Config
def write_to_xyz(vaspxml_path, save_path, Config_type, append=True):

    atoms_list = []
    atoms = ase_read(vaspxml_path, index=":")
    index = 1
    for atom in atoms:
        xx, yy, zz, yz, xz, xy = - atom.calc.results['stress'] * atom.get_volume()  # *160.21766
        atom.info['virial'] = np.array([(xx, xy, xz), (xy, yy, yz), (xz, yz, zz)])

        atom.calc.results['energy'] = atom.calc.results['free_energy']

        atom.info['Config_type'] = Config_type + str(index)

        atom.info['Weight'] = 1.0

        del atom.calc.results['stress']
        del atom.calc.results['free_energy']
        atoms_list.append(atom)
        index += 1

    ase_write(save_path, atoms_list, format='extxyz', append=append)
    return atoms_list
class VaspInput(Vasp):


    def __init__(self,*args,**kwargs):

        super(VaspInput,self).__init__(*args,**kwargs)
        self.input_params["setups"] = {"base": "recommended"}
        self.input_params["pp"] = ''
        
        os.environ[self.VASP_PP_PATH] = os.path.expanduser(Config.get("environ", "potcar_path"))

    def calculate(self,
                  atoms=None,
                  properties=('energy', ),
                  system_changes=tuple(calculator.all_changes)):
        """Do a VASP calculation in the specified directory.

        This will generate the necessary VASP input files, and then
        execute VASP. After execution, the energy, forces. etc. are read
        from the VASP output files.
        """

                      
        Calculator.calculate(self, atoms, properties, system_changes)
        # Check for zero-length lattice vectors and PBC
        # and that we actually have an Atoms object.
        check_atoms(self.atoms)

        self.clear_results()

        command = self.make_command(self.command)
        self.write_input(self.atoms, properties, system_changes)

        with self._txt_outstream() as out:
            errorcode, stderr = self._run(command=command,
                                          out=out,
                                          directory=self.directory)

        if errorcode:
            raise calculator.CalculationFailed(
                '{} in {} returned an error: {:d} stderr {}'.format(
                    self.name, Path(self.directory).resolve(), errorcode,
                    stderr))

        # Read results from calculation
        self.update_atoms(atoms)
        self.read_results()
    def _run(self, command=None, out=None, directory=None):
        """Method to explicitly execute VASP"""
        if command is None:
            command = self.command
        if directory is None:
            directory = self.directory

        result = subprocess.run(command,
                                shell=True,
                                cwd=directory,
                                capture_output=True,
                                text=True)
        if out is not None:
            out.write(result.stdout)
            out.write(result.stderr)

        return result.returncode, result.stderr


if __name__ == '__main__':
    vasp=VaspInput()




    atoms=ase_read("./POSCAR",format='vasp')
    vasp.read_incar("./INCAR")
    vasp.calculate(atoms,('energy'))
    print(vasp.results)
    print(vasp.atoms.info)
    print(atoms.calc.results)
